# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utilities for interfacing dm_env and dm_env_rpc."""

from typing import Dict
from dm_env import specs
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import spec_manager as dm_env_rpc_spec_manager
from dm_env_rpc.v1 import tensor_spec_utils
from dm_env_rpc.v1 import tensor_utils


def tensor_spec_to_dm_env_spec(
    tensor_spec: dm_env_rpc_pb2.TensorSpec) -> specs.Array:
  """Returns a dm_env spec given a dm_env_rpc TensorSpec.

  Args:
    tensor_spec: A dm_env_rpc TensorSpec protobuf.

  Returns:
    Either a DiscreteArray, BoundedArray, StringArray or Array, depending on the
    content of the TensorSpec.
  """
  np_type = tensor_utils.data_type_to_np_type(tensor_spec.dtype)
  if tensor_spec.HasField('min') or tensor_spec.HasField('max'):
    bounds = tensor_spec_utils.bounds(tensor_spec)

    if (not tensor_spec.shape
        and np.issubdtype(np_type, np.integer)
        and bounds.min == 0
        and tensor_spec.HasField('max')):
      return specs.DiscreteArray(
          num_values=bounds.max + 1, dtype=np_type, name=tensor_spec.name)
    else:
      return specs.BoundedArray(
          shape=tensor_spec.shape,
          dtype=np_type,
          name=tensor_spec.name,
          minimum=bounds.min,
          maximum=bounds.max)
  else:
    if tensor_spec.dtype == dm_env_rpc_pb2.DataType.STRING:
      return specs.StringArray(shape=tensor_spec.shape, name=tensor_spec.name)
    else:
      return specs.Array(
          shape=tensor_spec.shape, dtype=np_type, name=tensor_spec.name)


def dm_env_spec(
    spec_manager: dm_env_rpc_spec_manager.SpecManager
) -> Dict[str, specs.Array]:
  """Returns a dm_env spec for the given `spec_manager`.

  Args:
    spec_manager: An instance of SpecManager.

  Returns:
    A dict mapping names to either a dm_env Array, BoundedArray, DiscreteArray
    or StringArray spec for each named TensorSpec in `spec_manager`.
  """
  return {
      name: tensor_spec_to_dm_env_spec(spec_manager.name_to_spec(name))
      for name in spec_manager.names()
  }



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests Properties extension."""

import contextlib

from absl.testing import absltest
from dm_env import specs
import mock
import numpy as np

from google.protobuf import any_pb2
from google.rpc import code_pb2
from google.rpc import status_pb2
from google.protobuf import text_format
from dm_env_rpc.v1 import connection as dm_env_rpc_connection
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1.extensions import properties
from dm_env_rpc.v1.extensions import properties_pb2


def _create_property_request_key(text_proto):
  extension_message = any_pb2.Any()
  extension_message.Pack(
      text_format.Parse(text_proto, properties_pb2.PropertyRequest()))
  return dm_env_rpc_pb2.EnvironmentRequest(
      extension=extension_message).SerializeToString()


def _pack_property_response(text_proto):
  extension_message = any_pb2.Any()
  extension_message.Pack(
      text_format.Parse(text_proto, properties_pb2.PropertyResponse()))
  return dm_env_rpc_pb2.EnvironmentResponse(extension=extension_message)

# Set of expected requests and associated responses for mock connection.
_EXPECTED_REQUEST_RESPONSE_PAIRS = {
    _create_property_request_key('read_property { key: "foo" }'):
        _pack_property_response(
            'read_property { value: { int32s: { array: 1 } } }'),
    _create_property_request_key("""write_property {
               key: "bar"
               value: { strings { array: "some_value" } }
             }"""):
        _pack_property_response('write_property {}'),
    _create_property_request_key('read_property { key: "bar" }'):
        _pack_property_response(
            'read_property { value: { strings: { array: "some_value" } } }'),
    _create_property_request_key('list_property { key: "baz" }'):
        _pack_property_response("""list_property {
                    values: {
                      is_readable:true
                      spec { name: "baz.fiz" dtype:UINT32 shape: 2 shape: 2 }
                    }}"""),
    _create_property_request_key('list_property {}'):
        _pack_property_response("""list_property {
                 values: { is_readable:true spec { name: "foo" dtype:INT32 }
                           description: "This is a documented integer" }
                 values: { is_readable:true
                           is_writable:true
                           spec { name: "bar" dtype:STRING } }
                 values: { is_listable:true spec { name: "baz" } }
               }"""),
    _create_property_request_key('read_property { key: "bad_property" }'):
        dm_env_rpc_pb2.EnvironmentResponse(
            error=status_pb2.Status(message='invalid property request.')),
    _create_property_request_key('read_property { key: "invalid_key" }'):
        dm_env_rpc_pb2.EnvironmentResponse(
            error=status_pb2.Status(
                code=code_pb2.NOT_FOUND, message='Invalid key.')),
    _create_property_request_key("""write_property {
                                 key: "argument_test"
                                 value: { strings: { array: "invalid" } } }"""):
        dm_env_rpc_pb2.EnvironmentResponse(
            error=status_pb2.Status(
                code=code_pb2.INVALID_ARGUMENT, message='Invalid argument.')),
    _create_property_request_key('read_property { key: "permission_key" }'):
        dm_env_rpc_pb2.EnvironmentResponse(
            error=status_pb2.Status(
                code=code_pb2.PERMISSION_DENIED, message='No permission.'))
}


@contextlib.contextmanager
def _create_mock_connection():
  """Helper to create mock dm_env_rpc connection."""
  with mock.patch.object(dm_env_rpc_connection,
                         'dm_env_rpc_pb2_grpc') as mock_grpc:

    def _process(request_iterator, metadata):
      del metadata
      for request in request_iterator:
        yield _EXPECTED_REQUEST_RESPONSE_PAIRS[request.SerializeToString()]

    mock_stub_class = mock.MagicMock()
    mock_stub_class.Process = _process
    mock_grpc.EnvironmentStub.return_value = mock_stub_class
    yield dm_env_rpc_connection.Connection(mock.MagicMock())


class PropertiesTest(absltest.TestCase):

  def test_read_property(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      self.assertEqual(1, extension['foo'])

  def test_write_property(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      extension['bar'] = 'some_value'
      self.assertEqual('some_value', extension['bar'])

  def test_list_property(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      property_specs = extension.specs('baz')
      self.assertLen(property_specs, 1)

      property_spec = property_specs['baz.fiz']
      self.assertTrue(property_spec.readable)
      self.assertFalse(property_spec.writable)
      self.assertFalse(property_spec.listable)
      self.assertEqual(
          specs.Array(shape=(2, 2), dtype=np.uint32), property_spec.spec)

  def test_root_list_property(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      property_specs = extension.specs()
      self.assertLen(property_specs, 3)
      self.assertTrue(property_specs['foo'].readable)
      self.assertTrue(property_specs['bar'].readable)
      self.assertTrue(property_specs['bar'].writable)
      self.assertTrue(property_specs['baz'].listable)

  def test_invalid_spec_request_on_listable_property(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      property_specs = extension.specs()
      self.assertTrue(property_specs['baz'].listable)
      self.assertIsNone(property_specs['baz'].spec)

  def test_invalid_request(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      with self.assertRaisesRegex(error.DmEnvRpcError,
                                  'invalid property request.'):
        _ = extension['bad_property']

  def test_invalid_key_raises_key_error(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      with self.assertRaises(KeyError):
        _ = extension['invalid_key']

  def test_invalid_argument_raises_value_error(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      with self.assertRaises(ValueError):
        extension['argument_test'] = 'invalid'

  def test_permission_denied_raises_permission_error(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      with self.assertRaises(PermissionError):
        _ = extension['permission_key']

  def test_property_description(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      property_specs = extension.specs()
      self.assertEqual('This is a documented integer',
                       property_specs['foo'].description)

  def test_property_print(self):
    with _create_mock_connection() as connection:
      extension = properties.PropertiesExtension(connection)
      property_specs = extension.specs()
      self.assertRegex(
          str(property_specs['foo']),
          (r'PropertySpec\(key=foo, readable=True, writable=False, '
           r'listable=False, spec=.*, '
           r'description=This is a documented integer\)'))


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Manager class to manage the dm_env_rpc UID system."""

from typing import Any, Collection, Mapping, MutableMapping
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_utils


def _assert_shapes_match(tensor: dm_env_rpc_pb2.Tensor,
                         dm_env_rpc_spec: dm_env_rpc_pb2.TensorSpec):
  """Raises ValueError if shape of tensor and spec don't match."""
  tensor_shape = np.asarray(tensor.shape)
  spec_shape = np.asarray(dm_env_rpc_spec.shape)

  # Check all elements are equal, or the spec element is -1 (variable length).
  if tensor_shape.size != spec_shape.size or not np.all(
      (tensor_shape == spec_shape) | (spec_shape < 0)):
    raise ValueError(
        'Received dm_env_rpc tensor {} with shape {} but spec has shape {}.'
        .format(dm_env_rpc_spec.name, tensor_shape, spec_shape))


class SpecManager(object):
  """Manages transitions between Python dicts and dm_env_rpc UIDs.

  To make sending and receiving actions and observations easier for dm_env_rpc,
  this helps manage the transition between UID-keyed dicts mapping to dm_env_rpc
  tensors and string-keyed dicts mapping to scalars, lists, or NumPy arrays.
  """

  def __init__(self, specs: Mapping[int, dm_env_rpc_pb2.TensorSpec]):
    """Builds the SpecManager from the given dm_env_rpc specs.

    Args:
      specs: A dict mapping UIDs to dm_env_rpc TensorSpecs, similar to what is
        stored in `actions` and `observations` in ActionObservationSpecs.
    """
    for spec in specs.values():
      if np.count_nonzero(np.asarray(spec.shape) < 0) > 1:
        raise ValueError(
            f'"{spec.name}" shape has > 1 variable length dimension. '
            f'Spec:\n{spec}')

    self._name_to_uid = {spec.name: uid for uid, spec in specs.items()}
    self._uid_to_name = {uid: spec.name for uid, spec in specs.items()}
    if len(self._name_to_uid) != len(self._uid_to_name):
      raise ValueError('There are duplicate names in the tensor specs.')

    self._specs_by_uid = specs
    self._specs_by_name = {spec.name: spec for spec in specs.values()}

  @property
  def specs_by_uid(self) -> MutableMapping[int, dm_env_rpc_pb2.TensorSpec]:
    return self._specs_by_uid

  @property
  def specs_by_name(self) -> MutableMapping[str, dm_env_rpc_pb2.TensorSpec]:
    return self._specs_by_name

  def name_to_uid(self, name: str) -> int:
    """Returns the UID for the given name."""
    return self._name_to_uid[name]

  def uid_to_name(self, uid: int) -> str:
    """Returns the name for the given UID."""
    return self._uid_to_name[uid]

  def name_to_spec(self, name: str) -> dm_env_rpc_pb2.TensorSpec:
    """Returns the dm_env_rpc TensorSpec named `name`."""
    return self._specs_by_name[name]

  def uid_to_spec(self, uid: int) -> dm_env_rpc_pb2.TensorSpec:
    """Returns the dm_env_rpc TensorSpec for the given UID."""
    return self._specs_by_uid[uid]

  def names(self) -> Collection[str]:
    """Returns the spec names in no particular order."""
    return self._name_to_uid.keys()

  def uids(self) -> Collection[int]:
    """Returns the spec UIDs in no particular order."""
    return self._uid_to_name.keys()

  def unpack(
      self, dm_env_rpc_tensors: Mapping[int, dm_env_rpc_pb2.Tensor]
  ) -> MutableMapping[str, Any]:
    """Unpacks a dm_env_rpc uid-to-tensor map to a name-keyed Python dict.

    Args:
      dm_env_rpc_tensors: A dict mapping UIDs to dm_env_rpc tensor protos.

    Returns:
      A dict mapping names to scalars and arrays.
    """
    unpacked = {}
    for uid, tensor in dm_env_rpc_tensors.items():
      name = self._uid_to_name[uid]
      dm_env_rpc_spec = self.name_to_spec(name)
      _assert_shapes_match(tensor, dm_env_rpc_spec)
      tensor_dtype = tensor_utils.get_tensor_type(tensor)
      spec_dtype = tensor_utils.data_type_to_np_type(dm_env_rpc_spec.dtype)
      if tensor_dtype != spec_dtype:
        raise ValueError(
            'Received dm_env_rpc tensor {} with dtype {} but spec has dtype {}.'
            .format(name, tensor_dtype, spec_dtype))
      tensor_unpacked = tensor_utils.unpack_tensor(tensor)
      unpacked[name] = tensor_unpacked
    return unpacked

  def pack(
      self,
      tensors: Mapping[str, Any]) -> MutableMapping[int, dm_env_rpc_pb2.Tensor]:
    """Packs a name-keyed Python dict to a dm_env_rpc uid-to-tensor map.

    Args:
      tensors: A dict mapping string names to scalars and arrays.

    Returns:
      A dict mapping UIDs to dm_env_rpc tensor protos.
    """
    packed = {}
    for name, value in tensors.items():
      dm_env_rpc_spec = self.name_to_spec(name)
      tensor = tensor_utils.pack_tensor(value, dtype=dm_env_rpc_spec.dtype)
      _assert_shapes_match(tensor, dm_env_rpc_spec)
      packed[self.name_to_uid(name)] = tensor
    return packed



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc helper functions."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_spec_utils


class NpRangeInfoTests(absltest.TestCase):

  def test_floating(self):
    expected_min = np.finfo(np.float32).min
    actual_min = tensor_spec_utils._np_range_info(np.float32).min
    self.assertEqual(expected_min, actual_min)

  def test_integer(self):
    actual_min = tensor_spec_utils._np_range_info(np.uint32).min
    self.assertEqual(0, actual_min)

  def test_string_gives_error(self):
    with self.assertRaisesRegex(ValueError, 'numpy.str_'):
      _ = tensor_spec_utils._np_range_info(np.str_).min


class BoundsTests(parameterized.TestCase):

  def test_unbounded_unsigned(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((0, 2**32 - 1), bounds)

  def test_unbounded_signed(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-2**31, 2**31 - 1), bounds)

  def test_min_n_shape(self):
    minimum = np.array([[1, 2], [3, 4]])

    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = minimum.flatten().data.tolist()
    tensor_spec.shape[:] = minimum.shape
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(minimum, bounds.min)
    self.assertEqual(2**32 - 1, bounds.max)

  def test_max_n_shape(self):
    maximum = np.array([[1, 2], [3, 4]])

    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = maximum.flatten().data.tolist()
    tensor_spec.shape[:] = maximum.shape
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual(0, bounds.min)
    np.testing.assert_array_equal(maximum, bounds.max)

  def test_invalid_min_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1, 2]
    with self.assertRaisesRegex(ValueError,
                                'Scalar tensors must have exactly 1 element.*'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_invalid_max_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1, 2]
    tensor_spec.shape[:] = (2, 2)
    with self.assertRaisesRegex(ValueError,
                                'cannot reshape array of size .* into shape.*'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((1, 2**32 - 1), bounds)

  def test_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((0, 1), bounds)

  def test_min_and_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1]
    tensor_spec.max.int32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-1, 1), bounds)

  def test_broadcast_var_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1]
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.shape[:] = (-1,)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-1, 1), bounds)

  def test_invalid_min_var_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1, -1]
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.shape[:] = (-1,)
    with self.assertRaisesRegex(
        ValueError, "TensorSpec's with variable length shapes "
        'can only have scalar ranges.'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min_scalar_doesnt_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.shape[:] = (2, 2)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual(1, bounds.min)
    self.assertEqual(2**32 - 1, bounds.max)

  def test_max_scalar_doesnt_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1]
    tensor_spec.shape[:] = (2, 2)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual(0, bounds.min)
    self.assertEqual(1, bounds.max)

  def test_min_max_scalars_dont_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.max.uint32s.array[:] = [2]
    tensor_spec.shape[:] = (4,)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual(1, bounds.min)
    self.assertEqual(2, bounds.max)

  def test_min_mismatches_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*uint32.*min.*int32'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_max_mismatches_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*uint32.*max.*int32'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_nonnumeric_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.STRING
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*non-numeric.*string'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_max_0_stays_0(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_spec.max.int8s.array = b'\x00'
    tensor_spec.name = 'foo'
    self.assertEqual((-128, 0), tensor_spec_utils.bounds(tensor_spec))

  def test_min_0_stays_0(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_spec.min.int8s.array = b'\x00'
    tensor_spec.name = 'foo'
    self.assertEqual((0, 127), tensor_spec_utils.bounds(tensor_spec))

  def test_max_less_than_min_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.max.int32s.array[:] = [-1]
    tensor_spec.min.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*min 1.*max -1'):
      tensor_spec_utils.bounds(tensor_spec)

  @parameterized.parameters([
      dict(minimum=0., maximum=np.inf),
      dict(minimum=-np.inf, maximum=0.),
      dict(minimum=-np.inf, maximum=np.inf),
  ])
  def test_infinite_bounds_are_valid_for_floats(self, minimum, maximum):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.DOUBLE
    tensor_spec.min.doubles.array[:] = [minimum]
    tensor_spec.max.doubles.array[:] = [maximum]
    tensor_spec.name = 'foo'
    tensor_spec_utils.bounds(tensor_spec)


class SetBoundsTests(parameterized.TestCase):

  def test_set_scalar_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=2)

    self.assertEqual([1], tensor_spec.min.int32s.array)
    self.assertEqual([2], tensor_spec.max.int32s.array)

  def test_set_scalar_bounds_int8(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.DataType.INT8)
    minimum = 1
    maximum = 2
    tensor_spec_utils.set_bounds(tensor_spec, minimum=minimum, maximum=maximum)
    self.assertEqual(np.int8(minimum).tobytes(), tensor_spec.min.int8s.array)
    self.assertEqual(np.int8(maximum).tobytes(), tensor_spec.max.int8s.array)

  def test_set_multiple_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=(2,), dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=[1, 2], maximum=[3, 4])

    self.assertEqual([1, 2], tensor_spec.min.int32s.array)
    self.assertEqual([3, 4], tensor_spec.max.int32s.array)

  def test_set_only_min(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=None)

    self.assertEqual([1], tensor_spec.min.int32s.array)
    self.assertIsNone(tensor_spec.max.WhichOneof('payload'))

  def test_set_only_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=1)

    self.assertIsNone(tensor_spec.min.WhichOneof('payload'))
    self.assertEqual([1], tensor_spec.max.int32s.array)

  def test_unset_min_and_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=2)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=None)

    self.assertIsNone(tensor_spec.min.WhichOneof('payload'))
    self.assertIsNone(tensor_spec.max.WhichOneof('payload'))

  def test_cannot_set_nonnumeric_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', dtype=dm_env_rpc_pb2.DataType.STRING)
    with self.assertRaisesRegex(ValueError, 'non-numeric'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=None)

  def test_can_set_broadcastable_min_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=[2, 3])

    self.assertEqual([1], tensor_spec.min.int32s.array)
    self.assertEqual([2, 3], tensor_spec.max.int32s.array)

  def test_cannot_set_multiple_min_bounds_on_variable_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2, -1], dtype=dm_env_rpc_pb2.DataType.INT32)
    with self.assertRaisesRegex(ValueError, 'incompatible'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=[2, 3], maximum=4)

  def test_cannot_set_dissimilar_shape_on_min_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2, 2], dtype=dm_env_rpc_pb2.DataType.INT32)
    with self.assertRaisesRegex(ValueError, 'incompatible'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=[1, 2, 3], maximum=4)

  def test_can_set_broadcastable_max_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
    tensor_spec_utils.set_bounds(tensor_spec, minimum=[2, 3], maximum=4)

    self.assertEqual([2, 3], tensor_spec.min.int32s.array)
    self.assertEqual([4], tensor_spec.max.int32s.array)

  def test_cannot_set_multiple_max_bounds_on_variable_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2, -1], dtype=dm_env_rpc_pb2.DataType.INT32)
    with self.assertRaisesRegex(ValueError, 'incompatible'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=[2, 3])

  def test_cannot_set_dissimilar_shape_on_max_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2, 2], dtype=dm_env_rpc_pb2.DataType.INT32)
    with self.assertRaisesRegex(ValueError, 'incompatible'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=0, maximum=[1, 2, 3])

  def test_cannot_set_any_min_bounds_to_exceed_maximum(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec(
        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
    with self.assertRaisesRegex(ValueError, 'larger than max'):
      tensor_spec_utils.set_bounds(tensor_spec, minimum=[0, 4], maximum=[1, 1])

  @parameterized.parameters([
      dict(minimum=[-1000], maximum=[1]),
      dict(minimum=[1], maximum=[1000]),
  ])
  def test_new_bounds_must_be_safely_castable_to_dtype(self, minimum, maximum):
    name = 'test'
    dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_spec = dm_env_rpc_pb2.TensorSpec(name=name, dtype=dtype)
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        tensor_spec_utils._BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE.format(
            name=name,
            minimum=minimum,
            maximum=maximum,
            dtype=dm_env_rpc_pb2.DataType.Name(dtype))):
      tensor_spec_utils.set_bounds(
          tensor_spec, minimum=minimum, maximum=maximum)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc/message_utils."""

from absl.testing import absltest

from google.rpc import status_pb2
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import message_utils
from dm_env_rpc.v1 import tensor_utils


_CREATE_WORLD_REQUEST = dm_env_rpc_pb2.CreateWorldRequest(
    settings={'foo': tensor_utils.pack_tensor('bar')})
_CREATE_WORLD_RESPONSE = dm_env_rpc_pb2.CreateWorldResponse(world_name='qux')
_CREATE_WORLD_ENVIRONMENT_RESPONSE = dm_env_rpc_pb2.EnvironmentResponse(
    create_world=_CREATE_WORLD_RESPONSE)


class MessageUtilsTests(absltest.TestCase):

  def test_pack_create_world_request(self):
    environment_request, field_name = message_utils.pack_environment_request(
        _CREATE_WORLD_REQUEST)
    self.assertEqual(field_name, 'create_world')
    self.assertEqual(environment_request.WhichOneof('payload'),
                     'create_world')
    self.assertEqual(environment_request.create_world,
                     _CREATE_WORLD_REQUEST)

  def test_unpack_create_world_response(self):
    response = message_utils.unpack_environment_response(
        _CREATE_WORLD_ENVIRONMENT_RESPONSE, 'create_world')
    self.assertEqual(response, _CREATE_WORLD_RESPONSE)

  def test_unpack_error_response(self):
    with self.assertRaisesRegex(error.DmEnvRpcError, 'A test error.'):
      message_utils.unpack_environment_response(
          dm_env_rpc_pb2.EnvironmentResponse(
              error=status_pb2.Status(message='A test error.')),
          'create_world')

  def test_unpack_incorrect_response(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        'Unexpected response message! expected: create_world, actual: '
        'leave_world'):
      message_utils.unpack_environment_response(
          dm_env_rpc_pb2.EnvironmentResponse(
              leave_world=dm_env_rpc_pb2.LeaveWorldResponse()),
          'create_world')


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper Python utilities for bridging dm_env_rpc and NumPy.

Note that the Tensor proto payload type is not supported, as it doesn't play
well with NumPy.
"""
import abc
from typing import Union
import numpy as np

from google.protobuf import any_pb2
from google.protobuf import message
from dm_env_rpc.v1 import dm_env_rpc_pb2


def _pack_any_proto(value):
  """Helper function to pack Any proto, iff it's not already packed."""
  if isinstance(value, any_pb2.Any):
    return value
  elif isinstance(value, message.Message):
    any_proto = any_pb2.Any()
    any_proto.Pack(value)
    return any_proto
  else:
    # If we reach this exception, it is normally because the type being packed
    # is not supported. Raise exception with some typical examples.
    raise ValueError("Trying to pack an Any proto with a type that's not "
                     f"recognized! Type: {type(value)}, value: '{value}'. "
                     'Is the value a jagged iterable? Is the data type not a '
                     'supported primitive type like strings, floats, integers '
                     'or protobuf messages? Are all elements in the array the '
                     'same type?')


TensorOrTensorSpecValue = Union[dm_env_rpc_pb2.Tensor,
                                dm_env_rpc_pb2.TensorSpec.Value]


class Packer(metaclass=abc.ABCMeta):
  """Converts between proto messages and NumPy arrays."""

  def __init__(self, name: str, np_type: np.dtype):
    self._np_type = np_type
    self._name = name

  @property
  def name(self) -> str:
    return self._name

  @property
  def np_type(self) -> np.dtype:
    return self._np_type

  @abc.abstractmethod
  def pack(self, proto: TensorOrTensorSpecValue, value: np.ndarray):
    """Flattens and stores the given `value` array in the given `proto`."""

  @abc.abstractmethod
  def unpack(self, proto: TensorOrTensorSpecValue) -> np.ndarray:
    """Retrieves a flat NumPy array for the payload from the given `proto`."""


class _RepeatedFieldPacker(Packer):
  """Handles packing and unpacking most data types."""

  def pack(self, proto: TensorOrTensorSpecValue, value: np.ndarray):
    payload = getattr(proto, self._name)
    payload.array.extend(value.ravel().tolist())

  def unpack(self, proto: TensorOrTensorSpecValue) -> np.ndarray:
    payload = getattr(proto, self._name)
    return np.fromiter(payload.array, self.np_type, len(payload.array))


class _BytesPacker(Packer):
  """Handles packing and unpacking int8 and uint8 arrays."""

  def pack(self, proto: TensorOrTensorSpecValue, value: np.ndarray):
    payload = getattr(proto, self.name)
    payload.array = value.tobytes()

  def unpack(self, proto: TensorOrTensorSpecValue) -> np.ndarray:
    payload = getattr(proto, self.name)
    return np.frombuffer(payload.array, self.np_type)


class _RepeatedStringFieldPacker(Packer):
  """Handles packing and unpacking strings."""

  def pack(self, proto: TensorOrTensorSpecValue, value: np.ndarray):
    payload = getattr(proto, self.name)
    payload.array.extend(value.ravel().tolist())

  def unpack(self, proto: TensorOrTensorSpecValue) -> np.ndarray:
    # String arrays with variable length strings can't be created with
    # np.fromiter, unlike other dtypes.
    payload = getattr(proto, self.name)
    return np.array(payload.array, self.np_type)


class _RepeatedProtoFieldPacker(Packer):
  """Handles packing of protos."""

  def pack(self, proto: TensorOrTensorSpecValue, value: np.ndarray):
    payload = getattr(proto, self.name)
    payload.array.extend(
        [_pack_any_proto(sub_value) for sub_value in value.ravel()])

  def unpack(self, proto: TensorOrTensorSpecValue) -> np.ndarray:
    payload = getattr(proto, self._name)
    return np.array(payload.array, self.np_type)


_PACKERS = (
    _RepeatedFieldPacker('floats', np.float32),
    _RepeatedFieldPacker('doubles', np.float64),
    _BytesPacker('int8s', np.int8),
    _RepeatedFieldPacker('int32s', np.int32),
    _RepeatedFieldPacker('int64s', np.int64),
    _BytesPacker('uint8s', np.uint8),
    _RepeatedFieldPacker('uint32s', np.uint32),
    _RepeatedFieldPacker('uint64s', np.uint64),
    _RepeatedFieldPacker('bools', np.bool_),
    _RepeatedStringFieldPacker('strings', np.str_),
    _RepeatedProtoFieldPacker('protos', np.object_),
)

_NAME_TO_NP_TYPE = {
    packer.name: packer.np_type for packer in _PACKERS
}

_TYPE_TO_PACKER = {
    packer.np_type: packer for packer in _PACKERS
}

_DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE = {
    dm_env_rpc_pb2.DataType.FLOAT: np.float32,
    dm_env_rpc_pb2.DataType.DOUBLE: np.float64,
    dm_env_rpc_pb2.DataType.INT8: np.int8,
    dm_env_rpc_pb2.DataType.INT32: np.int32,
    dm_env_rpc_pb2.DataType.INT64: np.int64,
    dm_env_rpc_pb2.DataType.UINT8: np.uint8,
    dm_env_rpc_pb2.DataType.UINT32: np.uint32,
    dm_env_rpc_pb2.DataType.UINT64: np.uint64,
    dm_env_rpc_pb2.DataType.BOOL: np.bool_,
    dm_env_rpc_pb2.DataType.STRING: np.str_,
    dm_env_rpc_pb2.DataType.PROTO: np.object_,
}

_NUMPY_DTYPE_TO_DM_ENV_RPC_DTYPE = {
    value: key
    for key, value in _DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.items()
}


def get_tensor_type(tensor_proto: dm_env_rpc_pb2.Tensor) -> np.dtype:
  """Returns the NumPy type for the given tensor."""
  payload = tensor_proto.WhichOneof('payload')
  np_type = _NAME_TO_NP_TYPE.get(payload)
  if not np_type:
    raise TypeError(f'Unknown NumPy type {payload}')
  return np_type


def data_type_to_np_type(dm_env_rpc_dtype: dm_env_rpc_pb2.DataType) -> np.dtype:
  """Returns the NumPy type for the given dm_env_rpc DataType."""
  np_type = _DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.get(dm_env_rpc_dtype)
  if not np_type:
    raise TypeError(f'Unknown DataType {dm_env_rpc_dtype}')
  return np_type


def np_type_to_data_type(np_type: np.dtype) -> dm_env_rpc_pb2.DataType:
  """Returns the dm_env_rpc DataType for the given NumPy type."""
  if isinstance(np_type, np.dtype):
    # Flatten scalar types, since np.int32 is different from np.dtype(np.int32)
    # for dict key lookup.
    np_type = np_type.type
  data_type = _NUMPY_DTYPE_TO_DM_ENV_RPC_DTYPE.get(np_type)
  if data_type is None:
    raise TypeError(
        f'No dm_env_rpc DataType corresponds to NumPy type "{np_type}"')
  return data_type


def get_packer(np_type: np.dtype) -> Packer:
  """Retrieves the `Packer` which can handle the given NumPy Type.

  Note: The returned packer is a relatively low level mechanism to convert
  between NumPy arrays and the repeated `payload` fields in dm_env_rpc messages.
  It won't set shape or type on the proto message.  Instead of this packer,
  generally you should use `pack_tensor` and `unpack_tensor` to pack and unpack
  data to `Tensor` messages, as it will handle setting shape and type on the
  `Tensor` message as well.

  Args:
    np_type: The NumPy data type to retrieve a packer for.  eg: np.int32.

  Returns:
    An instance of Packer which will handle conversion between NumPy arrays of
    `np_type` and the corresponding payload field in the dm_env_rpc message.

  Raises:
    TypeError: If the provided NumPy type has no known packer.
  """
  packer = _TYPE_TO_PACKER.get(np_type)
  if not packer:
    raise TypeError(f'Unknown NumPy type "{np_type}" has no known packer.')
  return packer


def reshape_array(array: np.ndarray, shape):
  """Reshapes `array` to the given `shape` using dm_env_rpc's rules."""
  if shape:
    if len(array) == 1:
      array = np.full(np.maximum(shape, 1), array[0])
    else:
      array.shape = shape
    return array
  else:
    length = len(array)
    if length != 1:
      raise ValueError(
          'Scalar tensors must have exactly 1 element but had {} elements.'
          .format(length))
    return array[0]


def unpack_proto(proto: TensorOrTensorSpecValue) -> np.ndarray:
  """Converts a proto with payload oneof to a scalar or NumPy array.

  Args:
    proto: A dm_env_rpc proto with payload oneof.

  Returns:
    Returns a NumPy array of the payload with the correct type.
  """
  np_type = get_tensor_type(proto)
  packer = get_packer(np_type)
  return packer.unpack(proto)


def unpack_tensor(tensor_proto: dm_env_rpc_pb2.Tensor):
  """Converts a Tensor proto to a scalar or NumPy array.

  Args:
    tensor_proto: A dm_env_rpc Tensor protobuf.

  Returns:
    If the provided tensor_proto has a non-empty `shape` attribute, returns
    a NumPy array of the payload with the correct type and shape.  If the
    `shape` attribute is empty, returns a scalar (float, int, string, etc.)
    of the correct type and value.
  """
  array = unpack_proto(tensor_proto)
  return reshape_array(array, tensor_proto.shape)


def pack_tensor(value, dtype=None, try_compress=False) -> dm_env_rpc_pb2.Tensor:
  """Encodes the given value as a tensor.

  Args:
    value: A scalar (float, int, string, etc.), protobuf message, NumPy array,
      or nested lists.
    dtype: The type to pack the data to.  If set to None, will attempt to detect
      the correct type automatically.  Either a dm_env_rpc DataType enum or
      NumPy type is acceptable.
    try_compress: A bool, whether to try and encode the tensor in less space or
      not. This will increase the computational cost of the packing, but may
      reduce the on-the-wire size of the tensor.  There are no guarantees that
      any compression will actually happen.

  Raises:
    ValueError: If `value` is a jagged array, not a primitive type, nested
      iterable of primitive types or protobuf messages, or all elements can't be
      cast to the same type or the requested type.

  Returns:
    A dm_env_rpc Tensor proto containing the data.
  """
  packed = dm_env_rpc_pb2.Tensor()
  value = np.asarray(value)

  if value.dtype == np.object:
    # Because Numpy doesn't truly support variable-length string arrays, users
    # tend to use arrays of Numpy objects instead. Iff a user provides an array
    # of objects and a string dtype argument, automatically convert the value to
    # an array of strings.
    if np.issubdtype(
        _DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.get(dtype, dtype), np.str_):
      for item in value.flat:
        if not isinstance(item, str):
          raise TypeError(f'Requested string dtype but not all elements are '
                          'Python string types. At least one element was '
                          f'{type(item)}.')
      value = np.array(value, dtype=np.str_)

  elif dtype is not None:
    value = value.astype(
        dtype=_DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.get(dtype, dtype),
        copy=False,
        casting='same_kind')

  packed.shape[:] = value.shape
  packer = _TYPE_TO_PACKER[value.dtype.type]
  if (try_compress and np.all(value == next(value.flat))):
    # All elements are the same.  Pack in to a single value.
    packer.pack(packed, value.ravel()[0:1])
  else:
    packer.pack(packed, value)
  return packed



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for environment_stream.proto.

These aren't for testing functionality (it's assumed protobufs work) but for
testing/demonstrating how the protobufs would have to be used in code.
"""

from absl.testing import absltest
from dm_env_rpc.v1 import dm_env_rpc_pb2


class TensorTests(absltest.TestCase):

  def test_setting_tensor_data(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1, 2]

  def test_setting_tensor_data_with_wrong_type(self):
    tensor = dm_env_rpc_pb2.Tensor()
    with self.assertRaises(TypeError):
      tensor.floats.array[:] = ['hello!']

  def test_which_is_set(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1, 2]
    self.assertEqual('floats', tensor.WhichOneof('payload'))


class TensorSpec(absltest.TestCase):

  def test_setting_spec(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.name = 'Foo'
    tensor_spec.min.floats.array[:] = [0.0]
    tensor_spec.max.floats.array[:] = [0.0]
    tensor_spec.shape[:] = [2, 2]
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.FLOAT


class JoinWorldResponse(absltest.TestCase):

  def test_setting_spec(self):
    response = dm_env_rpc_pb2.JoinWorldResponse()
    tensor_spec = response.specs.actions[1]
    tensor_spec.shape[:] = [1]
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.FLOAT


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Compliance test base classes for dm_env_rpc."""
from dm_env_rpc.v1.compliance import create_destroy_world
from dm_env_rpc.v1.compliance import join_leave_world
from dm_env_rpc.v1.compliance import reset
from dm_env_rpc.v1.compliance import reset_world
from dm_env_rpc.v1.compliance import step

CreateDestroyWorld = create_destroy_world.CreateDestroyWorld
JoinLeaveWorld = join_leave_world.JoinLeaveWorld
Reset = reset.Reset
ResetWorld = reset_world.ResetWorld
Step = step.Step



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper Python utilities for managing dm_env_rpc TensorSpecs."""

import collections
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_utils


_BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE = (
    'TensorSpec "{name}"\'s bounds [{minimum}, {maximum}] contain value(s) '
    'that cannot be safely cast to dtype {dtype}.')

Bounds = collections.namedtuple('Bounds', ['min', 'max'])


def _can_cast(array_or_scalar, np_dtype: np.dtype) -> bool:
  for value in np.asarray(array_or_scalar).flat:
    if not np.can_cast(value, np_dtype, casting='safe'):
      return False
  return True


def _np_range_info(np_type: np.dtype):
  """Returns type info for `np_type`, which includes min and max attributes."""
  if issubclass(np_type, np.floating):
    return np.finfo(np_type)
  elif issubclass(np_type, np.integer):
    return np.iinfo(np_type)
  else:
    raise ValueError('{} does not have range info.'.format(np_type))


def _get_value(min_max_value, shape, default):
  """Helper function that returns the min/max bounds for a Value message.

  Args:
    min_max_value: Value protobuf message to get value from.
    shape: Optional dimensions to unpack payload data to.
    default: Value to use if min_max_value is not set.

  Returns:
    A scalar if `shape` is empty or None, or an unpacked NumPy array of either
    the unpacked value or provided default.

  """
  which = min_max_value.WhichOneof('payload')
  value = which and getattr(min_max_value, which)

  if value is None:
    min_max = default
  else:
    unpacked = tensor_utils.unpack_proto(min_max_value)
    min_max = tensor_utils.reshape_array(
        unpacked, shape) if len(unpacked) > 1 else unpacked[0]

  if (shape is not None
      and np.any(np.array(shape) < 0)
      and np.asarray(min_max).size > 1):
    raise ValueError(
        "TensorSpec's with variable length shapes can only have scalar ranges. "
        'Shape: {}, value: {}'.format(shape, min_max))
  return min_max


def bounds(tensor_spec: dm_env_rpc_pb2.TensorSpec) -> Bounds:
  """Gets the inclusive bounds of `tensor_spec`.

  Args:
    tensor_spec: An instance of a dm_env_rpc TensorSpec proto.

  Returns:
    A named tuple (`min`, `max`) of inclusive bounds.

  Raises:
    ValueError: `tensor_spec` does not have a numeric dtype, or the type of its
      `min` or `max` does not match its dtype, or the the bounds are invalid in
      some way.
  """
  np_type = tensor_utils.data_type_to_np_type(tensor_spec.dtype)
  tensor_spec_type = dm_env_rpc_pb2.DataType.Name(tensor_spec.dtype).lower()
  if not issubclass(np_type, np.number):
    raise ValueError('TensorSpec "{}" has non-numeric type {}.'
                     .format(tensor_spec.name, tensor_spec_type))

  # Check min payload type matches the tensor type.
  min_which = tensor_spec.min.WhichOneof('payload')
  if min_which and not min_which.startswith(tensor_spec_type):
    raise ValueError('TensorSpec "{}" has dtype {} but min type {}.'.format(
        tensor_spec.name, tensor_spec_type, min_which))

  # Check max payload type matches the tensor type.
  max_which = tensor_spec.max.WhichOneof('payload')
  if max_which and not max_which.startswith(tensor_spec_type):
    raise ValueError('TensorSpec "{}" has dtype {} but max type {}.'.format(
        tensor_spec.name, tensor_spec_type, max_which))

  dtype_bounds = _np_range_info(np_type)
  min_bound = _get_value(tensor_spec.min, tensor_spec.shape, dtype_bounds.min)
  max_bound = _get_value(tensor_spec.max, tensor_spec.shape, dtype_bounds.max)

  if not _can_cast(min_bound, np_type) or not _can_cast(max_bound, np_type):
    raise ValueError(
        _BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE.format(
            name=tensor_spec.name,
            minimum=min_bound,
            maximum=max_bound,
            dtype=tensor_spec_type))

  if np.any(max_bound < min_bound):
    raise ValueError('TensorSpec "{}" has min {} larger than max {}.'.format(
        tensor_spec.name, min_bound, max_bound))

  return Bounds(np_type(min_bound), np_type(max_bound))


def set_bounds(tensor_spec: dm_env_rpc_pb2.TensorSpec, minimum, maximum):
  """Modifies `tensor_spec` to have its inclusive bounds set.

  Packs `minimum` in to `tensor_spec.min` and `maximum` in to `tensor_spec.max`.

  Args:
    tensor_spec: An instance of a dm_env_rpc TensorSpec proto.  It should
      already have its `name`, `dtype` and `shape` attributes set.
    minimum: The minimum value that elements in the described tensor can obtain.
      A scalar, iterable of scalars, or None.  If None, `min` will be cleared on
      `tensor_spec`.
    maximum: The maximum value that elements in the described tensor can obtain.
      A scalar, iterable of scalars, or None.  If None, `max` will be cleared on
      `tensor_spec`.
  """
  np_type = tensor_utils.data_type_to_np_type(tensor_spec.dtype)
  if not issubclass(np_type, np.number):
    raise ValueError(f'TensorSpec has non-numeric type "{np_type}".')

  np_type_bounds = _np_range_info(np_type)
  has_min = minimum is not None
  has_max = maximum is not None

  if ((has_min and not _can_cast(minimum, np_type)) or
      (has_max and not _can_cast(maximum, np_type))):
    raise ValueError(
        _BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE.format(
            name=tensor_spec.name,
            minimum=minimum,
            maximum=maximum,
            dtype=dm_env_rpc_pb2.DataType.Name(tensor_spec.dtype)))

  if has_min:
    minimum = np.asarray(minimum, dtype=np_type)
    if minimum.size != 1 and minimum.shape != tuple(tensor_spec.shape):
      raise ValueError(
          f'minimum has shape {minimum.shape}, which is incompatible with '
          f"tensor_spec {tensor_spec.name}'s shape {tensor_spec.shape}.")

  if has_max:
    maximum = np.asarray(maximum, dtype=np_type)
    if maximum.size != 1 and maximum.shape != tuple(tensor_spec.shape):
      raise ValueError(
          f'maximum has shape {maximum.shape}, which is incompatible with '
          f"tensor_spec {tensor_spec.name}'s shape {tensor_spec.shape}.")

  if (has_min and has_max and np.any(maximum < minimum)):
    raise ValueError('TensorSpec "{}" has min {} larger than max {}.'.format(
        tensor_spec.name, minimum, maximum))

  packer = tensor_utils.get_packer(np_type)
  if has_min:
    packer.pack(tensor_spec.min, minimum)
  else:
    tensor_spec.ClearField('min')

  if has_max:
    packer.pack(tensor_spec.max, maximum)
  else:
    tensor_spec.ClearField('max')



# Lint as: python3
# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper functions used to process dm_env_rpc request / response messages.
"""

from typing import NamedTuple, Union

import immutabledict

from google.protobuf import any_pb2
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error


_MESSAGE_TYPE_TO_FIELD = immutabledict.immutabledict({
    field.message_type.name: field.name
    for field in dm_env_rpc_pb2.EnvironmentRequest.DESCRIPTOR.fields
})

DmEnvRpcRequest = Union[dm_env_rpc_pb2.CreateWorldRequest,
                        dm_env_rpc_pb2.JoinWorldRequest,
                        dm_env_rpc_pb2.StepRequest,
                        dm_env_rpc_pb2.ResetRequest,
                        dm_env_rpc_pb2.ResetWorldRequest,
                        dm_env_rpc_pb2.LeaveWorldRequest,
                        dm_env_rpc_pb2.DestroyWorldRequest,
                        any_pb2.Any,  # Extension message.
                       ]

DmEnvRpcResponse = Union[dm_env_rpc_pb2.CreateWorldResponse,
                         dm_env_rpc_pb2.JoinWorldResponse,
                         dm_env_rpc_pb2.StepResponse,
                         dm_env_rpc_pb2.ResetResponse,
                         dm_env_rpc_pb2.ResetWorldResponse,
                         dm_env_rpc_pb2.LeaveWorldResponse,
                         dm_env_rpc_pb2.DestroyWorldResponse,
                         any_pb2.Any,  # Extension message.
                        ]


class EnvironmentRequestAndFieldName(NamedTuple):
  """EnvironmentRequest and field name used when packing."""
  environment_request: dm_env_rpc_pb2.EnvironmentRequest
  field_name: str


def pack_environment_request(
    request: DmEnvRpcRequest) -> EnvironmentRequestAndFieldName:
  """Constructs an EnvironmentRequest containing a request message.

  Args:
    request: An instance of a dm_env_rpc Request type, such as
      CreateWorldRequest.

  Returns:
    A tuple of (environment_request, field_name) where:
      environment_request: dm_env_rpc.v1.EnvironmentRequest containing the input
        request message.
      field_name: Name of the environment request field holding the input
        request message.
  """
  field_name = _MESSAGE_TYPE_TO_FIELD[type(request).__name__]
  environment_request = dm_env_rpc_pb2.EnvironmentRequest()
  getattr(environment_request, field_name).CopyFrom(request)
  return EnvironmentRequestAndFieldName(environment_request, field_name)


def unpack_environment_response(
    environment_response: dm_env_rpc_pb2.EnvironmentResponse,
    expected_field_name: str) -> DmEnvRpcResponse:
  """Extracts the response message contained within an EnvironmentResponse.

  Args:
    environment_response: An instance of dm_env_rpc.v1.EnvironmentResponse.
    expected_field_name: Name of the environment response field expected to be
      holding the dm_env_rpc response message.

  Returns:
    dm_env_rpc response message wrapped in the input environment response.

  Raises:
    DmEnvRpcError: The dm_env_rpc EnvironmentResponse contains an error.
    ValueError: The dm_env_rpc response message contained in the
      EnvironmentResponse is held in a different field from the one expected.
  """
  response_field_name = environment_response.WhichOneof('payload')
  if response_field_name == 'error':
    raise error.DmEnvRpcError(environment_response.error)
  elif response_field_name == expected_field_name:
    return getattr(environment_response, expected_field_name)
  else:
    raise ValueError('Unexpected response message! expected: '
                     f'{expected_field_name}, actual: {response_field_name}')



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for SpecManager class."""

from absl.testing import absltest
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import spec_manager
from dm_env_rpc.v1 import tensor_utils

_EXAMPLE_SPECS = {
    54:
        dm_env_rpc_pb2.TensorSpec(
            name='fuzz', shape=[2], dtype=dm_env_rpc_pb2.DataType.FLOAT),
    55:
        dm_env_rpc_pb2.TensorSpec(
            name='foo', shape=[3], dtype=dm_env_rpc_pb2.DataType.INT32),
}


class SpecManagerTests(absltest.TestCase):

  def setUp(self):
    super(SpecManagerTests, self).setUp()
    self._spec_manager = spec_manager.SpecManager(_EXAMPLE_SPECS)

  def test_specs_by_uid(self):
    self.assertDictEqual(_EXAMPLE_SPECS, self._spec_manager.specs_by_uid)

  def test_specs_by_name(self):
    expected = {'foo': _EXAMPLE_SPECS[55], 'fuzz': _EXAMPLE_SPECS[54]}
    self.assertDictEqual(expected, self._spec_manager.specs_by_name)

  def test_name_to_uid(self):
    self.assertEqual(55, self._spec_manager.name_to_uid('foo'))

  def test_name_to_uid_no_such_name(self):
    with self.assertRaisesRegex(KeyError, 'bar'):
      self._spec_manager.name_to_uid('bar')

  def test_name_to_spec(self):
    spec = self._spec_manager.name_to_spec('foo')
    self.assertEqual([3], spec.shape)

  def test_name_to_spec_no_such_name(self):
    with self.assertRaisesRegex(KeyError, 'bar'):
      self._spec_manager.name_to_spec('bar')

  def test_uid_to_name(self):
    self.assertEqual('foo', self._spec_manager.uid_to_name(55))

  def test_uid_to_name_no_such_uid(self):
    with self.assertRaisesRegex(KeyError, '56'):
      self._spec_manager.uid_to_name(56)

  def test_names(self):
    self.assertEqual(set(['foo', 'fuzz']), self._spec_manager.names())

  def test_uids(self):
    self.assertEqual(set([54, 55]), self._spec_manager.uids())

  def test_uid_to_spec(self):
    spec = self._spec_manager.uid_to_spec(54)
    self.assertEqual([2], spec.shape)

  def test_pack(self):
    packed = self._spec_manager.pack({'fuzz': [1.0, 2.0], 'foo': [3, 4, 5]})
    expected = {
        54: tensor_utils.pack_tensor([1.0, 2.0], dtype=np.float32),
        55: tensor_utils.pack_tensor([3, 4, 5], dtype=np.int32),
    }
    self.assertDictEqual(expected, packed)

  def test_partial_pack(self):
    packed = self._spec_manager.pack({
        'fuzz': [1.0, 2.0],
    })
    expected = {
        54: tensor_utils.pack_tensor([1.0, 2.0], dtype=np.float32),
    }
    self.assertDictEqual(expected, packed)

  def test_pack_unknown_key_raises_error(self):
    with self.assertRaisesRegex(KeyError, 'buzz'):
      self._spec_manager.pack({'buzz': 'hello'})

  def test_pack_wrong_shape_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'shape'):
      self._spec_manager.pack({'foo': [1, 2]})

  def test_pack_wrong_dtype_raises_error(self):
    with self.assertRaisesRegex(TypeError, 'int32'):
      self._spec_manager.pack({'foo': 'hello'})

  def test_pack_cast_float_to_int_raises_error(self):
    with self.assertRaisesRegex(TypeError, 'int32'):
      self._spec_manager.pack({'foo': [0.5, 1.0, 1]})

  def test_pack_cast_int_to_float_is_ok(self):
    packed = self._spec_manager.pack({'fuzz': [1, 2]})
    self.assertEqual([1.0, 2.0], packed[54].floats.array)

  def test_unpack(self):
    unpacked = self._spec_manager.unpack({
        54: tensor_utils.pack_tensor([1.0, 2.0], dtype=np.float32),
        55: tensor_utils.pack_tensor([3, 4, 5], dtype=np.int32),
    })
    self.assertLen(unpacked, 2)
    np.testing.assert_array_equal(np.asarray([1.0, 2.0]), unpacked['fuzz'])
    np.testing.assert_array_equal(np.asarray([3, 4, 5]), unpacked['foo'])

  def test_partial_unpack(self):
    unpacked = self._spec_manager.unpack({
        54: tensor_utils.pack_tensor([1.0, 2.0], dtype=np.float32),
    })
    self.assertLen(unpacked, 1)
    np.testing.assert_array_equal(np.asarray([1.0, 2.0]), unpacked['fuzz'])

  def test_unpack_unknown_uid_raises_error(self):
    with self.assertRaisesRegex(KeyError, '53'):
      self._spec_manager.unpack({53: tensor_utils.pack_tensor('foo')})

  def test_unpack_wrong_shape_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'shape'):
      self._spec_manager.unpack({55: tensor_utils.pack_tensor([1, 2])})

  def test_unpack_wrong_type_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'dtype'):
      self._spec_manager.unpack(
          {55: tensor_utils.pack_tensor([1, 2, 3], dtype=np.float32)})


class SpecManagerVariableSpecShapeTests(absltest.TestCase):

  def setUp(self):
    super(SpecManagerVariableSpecShapeTests, self).setUp()
    specs = {
        101:
            dm_env_rpc_pb2.TensorSpec(
                name='foo', shape=[1, -1], dtype=dm_env_rpc_pb2.DataType.INT32),
    }
    self._spec_manager = spec_manager.SpecManager(specs)

  def test_variable_spec_shape(self):
    packed = self._spec_manager.pack({'foo': [[1, 2, 3, 4]]})
    expected = {
        101: tensor_utils.pack_tensor([[1, 2, 3, 4]], dtype=np.int32),
    }
    self.assertDictEqual(expected, packed)

  def test_invalid_variable_shape(self):
    with self.assertRaisesRegex(ValueError, 'shape'):
      self._spec_manager.pack({'foo': np.ones((1, 2, 3), dtype=np.int32)})

  def test_empty_variable_shape(self):
    manager = spec_manager.SpecManager({
        1:
            dm_env_rpc_pb2.TensorSpec(
                name='bar', shape=[], dtype=dm_env_rpc_pb2.DataType.INT32)
    })
    with self.assertRaisesRegex(ValueError, 'shape'):
      manager.pack({'bar': np.ones((1), dtype=np.int32)})

  def test_invalid_variable_spec_shape(self):
    with self.assertRaisesRegex(ValueError, 'shape has > 1 variable length'):
      spec_manager.SpecManager({
          1:
              dm_env_rpc_pb2.TensorSpec(
                  name='bar',
                  shape=[1, -1, -1],
                  dtype=dm_env_rpc_pb2.DataType.INT32)
      })


class SpecManagerConstructorTests(absltest.TestCase):

  def test_duplicate_names_raise_error(self):
    specs = {
        54:
            dm_env_rpc_pb2.TensorSpec(
                name='fuzz', shape=[3], dtype=dm_env_rpc_pb2.DataType.FLOAT),
        55:
            dm_env_rpc_pb2.TensorSpec(
                name='fuzz', shape=[2], dtype=dm_env_rpc_pb2.DataType.FLOAT),
    }
    with self.assertRaisesRegex(ValueError, 'duplicate name'):
      spec_manager.SpecManager(specs)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A helper class to manage a connection to a dm_env_rpc server.

This helper class allows sending the Request message types and receiving the
Response message types without wrapping in an EnvironmentRequest or unwrapping
from an EnvironmentResponse.  It also turns error messages in to exceptions.

For most calls (such as create, join, etc.):

    with connection.Connection(grpc_channel) as channel:
      create_response = channel.send(dm_env_rpc_pb2.CreateRequest(settings={
        'players': 5
      })

For the `extension` message type, you must send an Any proto and you'll get back
an Any proto.  It is up to you to wrap and unwrap these to concrete proto types
that you know how to handle.

    with connection.Connection(grpc_channel) as channel:
      request = struct_pb2.Struct()
      ...
      request_any = any_pb2.Any()
      request_any.Pack(request)
      response_any = channel.send(request_any)
      response = my_type_pb2.MyType()
      response_any.Unpack(response)


Any errors encountered in the EnvironmentResponse are turned into Python
exceptions, so explicit error handling code isn't needed per call.
"""

import queue
from typing import Optional, Sequence, Tuple
import grpc

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import dm_env_rpc_pb2_grpc
from dm_env_rpc.v1 import message_utils

Metadata = Sequence[Tuple[str, str]]


class _StreamReaderWriter(object):
  """Helper class for reading/writing gRPC streams."""

  def __init__(self,
               stub: dm_env_rpc_pb2_grpc.EnvironmentStub,
               metadata: Optional[Metadata] = None):
    self._requests = queue.Queue()
    self._stream = stub.Process(
        iter(self._requests.get, None), metadata=metadata)

  def write(self, request: dm_env_rpc_pb2.EnvironmentRequest):
    """Asynchronously sends `request` to the stream."""
    self._requests.put(request)

  def read(self) -> dm_env_rpc_pb2.EnvironmentResponse:
    """Returns the response from stream.  Blocking."""
    return next(self._stream)


class Connection(object):
  """A helper class for interacting with dm_env_rpc servers."""

  def __init__(self,
               channel: grpc.Channel,
               metadata: Optional[Metadata] = None):
    """Manages a connection to a dm_env_rpc server.

    Args:
      channel: A grpc channel to connect to the dm_env_rpc server over.
      metadata: Optional sequence of 2-tuples, sent to the gRPC server as
        metadata.
    """
    self._stream = _StreamReaderWriter(
        dm_env_rpc_pb2_grpc.EnvironmentStub(channel), metadata)

  def send(
      self,
      request: message_utils.DmEnvRpcRequest) -> message_utils.DmEnvRpcResponse:
    """Sends the given request to the dm_env_rpc server and returns the response.

    The request should be an instance of one of the dm_env_rpc Request messages,
    such as CreateWorldRequest. Based on the type the correct payload for the
    EnvironmentRequest will be constructed and sent to the dm_env_rpc server.

    Blocks until the server sends back its response.

    Args:
      request: An instance of a dm_env_rpc Request type, such as
        CreateWorldRequest.

    Returns:
      The response the dm_env_rpc server returned for the given RPC call,
      unwrapped from the EnvironmentStream message.  For instance if `request`
      had type `CreateWorldRequest` this returns a message of type
      `CreateWorldResponse`.

    Raises:
      DmEnvRpcError: The dm_env_rpc server responded to the request with an
        error.
      ValueError: The dm_env_rpc server responded to the request with an
        unexpected response message.
    """
    environment_request, field_name = (
        message_utils.pack_environment_request(request))
    self._stream.write(environment_request)
    return message_utils.unpack_environment_response(self._stream.read(),
                                                     field_name)

  def close(self):
    """Closes the connection.  Call when the connection is no longer needed."""
    if self._stream:
      self._stream = None

  def __exit__(self, *args, **kwargs):
    self.close()

  def __enter__(self):
    return self


def create_secure_channel_and_connect(
    server_address: str,
    credentials: grpc.ChannelCredentials = grpc.local_channel_credentials(),
    timeout: Optional[float] = None) -> Connection:
  """Creates a secure channel from server address and credentials and connects.

  We allow the created channel to have un-bounded message lengths, to support
  large observations.

  Args:
    server_address: URI server address to connect to.
    credentials: gRPC credentials necessary to connect to the server.
    timeout: Optional timeout in seconds to wait for channel to be ready.
      Default to waiting indefinitely.

  Returns:
    An instance of dm_env_rpc.Connection, where the channel is close upon the
    connection being closed.
  """
  options = [('grpc.max_send_message_length', -1),
             ('grpc.max_receive_message_length', -1)]
  channel = grpc.secure_channel(server_address, credentials, options=options)
  grpc.channel_ready_future(channel).result(timeout)

  class _ConnectionWrapper(Connection):
    """Utility to ensure channel is closed when the connection is closed."""

    def __init__(self, channel):
      super().__init__(channel)
      self._channel = channel

    def __del__(self):
      self.close()

    def close(self):
      super().close()
      self._channel.close()

  return _ConnectionWrapper(channel)



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Micro-benchmark for tensor_utils.pack_tensor."""

import abc
import timeit

from absl import app
from absl import flags
import numpy as np

from dm_env_rpc.v1 import tensor_utils

flags.DEFINE_integer('repeats', 10000,
                     'Number of times each benchmark will run.')
FLAGS = flags.FLAGS


class _AbstractBenchmark(metaclass=abc.ABCMeta):
  """Base class for benchmarks using timeit."""

  def run(self):
    time = timeit.timeit(self.statement, setup=self.setup, number=FLAGS.repeats)
    print(f'{self.name} -- overall: {time:0.2f} s, '
          f'per call: {time/FLAGS.repeats:0.1e} s')

  def setup(self):
    pass

  @abc.abstractmethod
  def statement(self):
    pass

  @abc.abstractproperty
  def name(self):
    pass


class _PackBenchmark(_AbstractBenchmark):
  """Benchmark for packing a numpy array to a Tensor proto."""

  def __init__(self, dtype, shape):
    self._name = f'pack {np.dtype(dtype).name}'
    self._dtype = dtype
    self._shape = shape

  @property
  def name(self):
    return self._name

  def setup(self):
    # Use non-zero values in case there's something special about zero arrays.
    self._unpacked = np.arange(
        np.prod(self._shape), dtype=self._dtype).reshape(self._shape)

  def statement(self):
    self._unpacked.flat[0] += 1  # prevent caching of the result
    tensor_utils.pack_tensor(self._unpacked, self._dtype)


class _UnpackBenchmark(_AbstractBenchmark):
  """Benchmark for unpacking a Tensor proto to a numpy array."""

  def __init__(self, dtype, shape):
    self._name = f'unpack {np.dtype(dtype).name}'
    self._shape = shape
    self._dtype = dtype

  @property
  def name(self):
    return self._name

  def setup(self):
    # Use non-zero values in case there's something special about zero arrays.
    tensor = np.arange(
        np.prod(self._shape), dtype=self._dtype).reshape(self._shape)
    self._packed = tensor_utils.pack_tensor(tensor, self._dtype)

  def statement(self):
    tensor_utils.unpack_tensor(self._packed)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  # Pick `shape` such that number of bytes is consistent between benchmarks.
  benchmarks = (
      _PackBenchmark(dtype=np.uint8, shape=(128, 128, 3)),
      _PackBenchmark(dtype=np.int32, shape=(64, 64, 3)),
      _PackBenchmark(dtype=np.int64, shape=(32, 64, 3)),
      _PackBenchmark(dtype=np.uint32, shape=(64, 64, 3)),
      _PackBenchmark(dtype=np.uint64, shape=(32, 64, 3)),
      _PackBenchmark(dtype=np.float32, shape=(64, 64, 3)),
      _PackBenchmark(dtype=np.float64, shape=(32, 64, 3)),
      _UnpackBenchmark(dtype=np.uint8, shape=(128, 128, 3)),
      _UnpackBenchmark(dtype=np.int32, shape=(64, 64, 3)),
      _UnpackBenchmark(dtype=np.int64, shape=(32, 64, 3)),
      _UnpackBenchmark(dtype=np.uint32, shape=(64, 64, 3)),
      _UnpackBenchmark(dtype=np.uint64, shape=(32, 64, 3)),
      _UnpackBenchmark(dtype=np.float32, shape=(64, 64, 3)),
      _UnpackBenchmark(dtype=np.float64, shape=(32, 64, 3)),
  )
  for benchmark in benchmarks:
    benchmark.run()


if __name__ == '__main__':
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for Connection."""

import contextlib

from absl.testing import absltest
import grpc
import mock

from google.protobuf import any_pb2
from google.protobuf import struct_pb2
from google.rpc import status_pb2
from dm_env_rpc.v1 import connection as dm_env_rpc_connection
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import tensor_utils

_CREATE_REQUEST = dm_env_rpc_pb2.CreateWorldRequest(
    settings={'foo': tensor_utils.pack_tensor('bar')})
_CREATE_RESPONSE = dm_env_rpc_pb2.CreateWorldResponse()

_BAD_CREATE_REQUEST = dm_env_rpc_pb2.CreateWorldRequest()
_TEST_ERROR = dm_env_rpc_pb2.EnvironmentResponse(
    error=status_pb2.Status(message='A test error.'))

_INCORRECT_RESPONSE_TEST_MSG = dm_env_rpc_pb2.DestroyWorldRequest(
    world_name='foo')
_INCORRECT_RESPONSE = dm_env_rpc_pb2.EnvironmentResponse(
    leave_world=dm_env_rpc_pb2.LeaveWorldResponse())

_EXTENSION_REQUEST = struct_pb2.Value(string_value='extension request')
_EXTENSION_RESPONSE = struct_pb2.Value(number_value=555)


def _wrap_in_any(proto):
  any_proto = any_pb2.Any()
  any_proto.Pack(proto)
  return any_proto


_REQUEST_RESPONSE_PAIRS = {
    dm_env_rpc_pb2.EnvironmentRequest(
        create_world=_CREATE_REQUEST).SerializeToString():
        dm_env_rpc_pb2.EnvironmentResponse(create_world=_CREATE_RESPONSE),
    dm_env_rpc_pb2.EnvironmentRequest(
        create_world=_BAD_CREATE_REQUEST).SerializeToString():
        _TEST_ERROR,
    dm_env_rpc_pb2.EnvironmentRequest(
        extension=_wrap_in_any(_EXTENSION_REQUEST)).SerializeToString():
        dm_env_rpc_pb2.EnvironmentResponse(
            extension=_wrap_in_any(_EXTENSION_RESPONSE)),
    dm_env_rpc_pb2.EnvironmentRequest(
        destroy_world=_INCORRECT_RESPONSE_TEST_MSG).SerializeToString():
        _INCORRECT_RESPONSE,
}


def _process(request_iterator, metadata):
  del metadata
  for request in request_iterator:
    yield _REQUEST_RESPONSE_PAIRS.get(request.SerializeToString(), _TEST_ERROR)


@contextlib.contextmanager
def _create_mock_channel():
  """Mocks out gRPC and returns a channel to be passed to Connection."""
  with mock.patch.object(dm_env_rpc_connection,
                         'dm_env_rpc_pb2_grpc') as mock_grpc:
    mock_stub_class = mock.MagicMock()
    mock_stub_class.Process = _process
    mock_grpc.EnvironmentStub.return_value = mock_stub_class
    yield mock.MagicMock()


class ConnectionTests(absltest.TestCase):

  def test_create(self):
    with _create_mock_channel() as mock_channel:
      with dm_env_rpc_connection.Connection(mock_channel) as connection:
        response = connection.send(_CREATE_REQUEST)
        self.assertEqual(_CREATE_RESPONSE, response)

  def test_error(self):
    with _create_mock_channel() as mock_channel:
      with dm_env_rpc_connection.Connection(mock_channel) as connection:
        with self.assertRaisesRegex(error.DmEnvRpcError, 'test error'):
          connection.send(_BAD_CREATE_REQUEST)

  def test_extension(self):
    with _create_mock_channel() as mock_channel:
      with dm_env_rpc_connection.Connection(mock_channel) as connection:
        request = any_pb2.Any()
        request.Pack(_EXTENSION_REQUEST)
        response = connection.send(request)
        expected_response = any_pb2.Any()
        expected_response.Pack(_EXTENSION_RESPONSE)
        self.assertEqual(expected_response, response)

  @mock.patch.object(grpc, 'secure_channel')
  @mock.patch.object(grpc, 'channel_ready_future')
  def test_create_secure_channel_and_connect(self, mock_channel_ready,
                                             mock_secure_channel):
    mock_channel = mock.MagicMock()
    mock_secure_channel.return_value = mock_channel

    self.assertIsNotNone(
        dm_env_rpc_connection.create_secure_channel_and_connect(
            'valid_address', grpc.local_channel_credentials()))

    mock_channel_ready.assert_called_once_with(mock_channel)
    mock_secure_channel.assert_called_once()
    mock_channel.close.assert_called_once()

  @mock.patch.object(grpc, 'secure_channel')
  @mock.patch.object(grpc, 'channel_ready_future')
  def test_create_secure_channel_and_connect_context(self, mock_channel_ready,
                                                     mock_secure_channel):
    mock_channel = mock.MagicMock()
    mock_secure_channel.return_value = mock_channel

    with dm_env_rpc_connection.create_secure_channel_and_connect(
        'valid_address') as connection:
      self.assertIsNotNone(connection)

    mock_channel_ready.assert_called_once_with(mock_channel)
    mock_secure_channel.assert_called_once()
    mock_channel.close.assert_called_once()

  def test_create_secure_channel_and_connect_timeout(self):
    with self.assertRaises(grpc.FutureTimeoutError):
      dm_env_rpc_connection.create_secure_channel_and_connect(
          'invalid_address', grpc.local_channel_credentials(), timeout=1.)

  def test_incorrect_response(self):
    with _create_mock_channel() as mock_channel:
      with dm_env_rpc_connection.Connection(mock_channel) as connection:
        with self.assertRaisesRegex(ValueError, 'Unexpected response message'):
          connection.send(_INCORRECT_RESPONSE_TEST_MSG)

  def test_with_metadata(self):
    expected_metadata = (('key', 'value'),)
    with mock.patch.object(dm_env_rpc_connection,
                           'dm_env_rpc_pb2_grpc') as mock_grpc:
      mock_stub_class = mock.MagicMock()
      mock_grpc.EnvironmentStub.return_value = mock_stub_class
      _ = dm_env_rpc_connection.Connection(
          mock.MagicMock(), metadata=expected_metadata)
      mock_stub_class.Process.assert_called_with(
          mock.ANY, metadata=expected_metadata)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc/dm_env adaptor."""

from absl.testing import absltest
import dm_env
from dm_env import specs
import mock
import numpy as np

from google.rpc import status_pb2
from google.protobuf import text_format
from dm_env_rpc.v1 import dm_env_adaptor
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import tensor_utils

_SAMPLE_STEP_REQUEST = dm_env_rpc_pb2.StepRequest(
    requested_observations=[1, 2],
    actions={
        1: tensor_utils.pack_tensor(4, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('hello')
    })
_SAMPLE_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })
_TERMINATED_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.TERMINATED,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })
_SAMPLE_SPEC = dm_env_rpc_pb2.ActionObservationSpecs(
    actions={
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo'),
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar')
    },
    observations={
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo'),
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar')
    },
)
_SAMPLE_SPEC_REORDERED = dm_env_rpc_pb2.ActionObservationSpecs(
    observations={
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar'),
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo')
    },
    actions={
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar'),
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo')
    },
)
_SAMPLE_NESTED_SPECS = dm_env_rpc_pb2.ActionObservationSpecs(
    actions={
        1:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.INT32, name='foo.bar'),
        2:
            dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='baz')
    },
    observations={
        1:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.INT32, name='foo.bar'),
        2:
            dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='baz')
    },
)

# Ensures the equality check in reset() works if the dictionary elements are
# created in a different order.
_SAMPLE_RESET_RESPONSE = dm_env_rpc_pb2.ResetResponse(
    specs=_SAMPLE_SPEC_REORDERED)
_RESET_CHANGES_SPEC_RESPONSE = dm_env_rpc_pb2.ResetResponse(
    specs=dm_env_rpc_pb2.ActionObservationSpecs())

_RESERVED_SPEC = dm_env_rpc_pb2.ActionObservationSpecs(
    actions={},
    observations={
        1:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.UINT8,
                name=dm_env_adaptor.DEFAULT_REWARD_KEY),
        2:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.STRING,
                name=dm_env_adaptor.DEFAULT_DISCOUNT_KEY)
    })
_RESERVED_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })


class DmEnvAdaptorTests(absltest.TestCase):

  def setUp(self):
    super(DmEnvAdaptorTests, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC)

  def test_requested_observations(self):
    requested_observations = ['foo']
    filtered_env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                               requested_observations)

    expected_filtered_step_request = dm_env_rpc_pb2.StepRequest(
        requested_observations=[1],
        actions={
            1: tensor_utils.pack_tensor(4, dtype=dm_env_rpc_pb2.UINT8),
            2: tensor_utils.pack_tensor('hello')
        })

    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    filtered_env.step({'foo': 4, 'bar': 'hello'})

    self._connection.send.assert_called_once_with(
        expected_filtered_step_request)

  def test_invalid_requested_observations(self):
    requested_observations = ['invalid']
    with self.assertRaisesRegex(ValueError,
                                'Unsupported observations requested'):
      dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                  requested_observations)

  def test_requested_observation_spec(self):
    requested_observations = ['foo']
    filtered_env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                               requested_observations)
    observation_names = [name for name in filtered_env.observation_spec()]
    self.assertEqual(requested_observations, observation_names)

  def test_first_running_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)
    self.assertIsNone(timestep.reward)
    self.assertIsNone(timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_mid_running_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self.assertEqual(dm_env.StepType.MID, timestep.step_type)
    self.assertEqual(0.0, timestep.reward)
    self.assertEqual(1.0, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_last_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        return_value=_TERMINATED_STEP_RESPONSE)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self.assertEqual(dm_env.StepType.LAST, timestep.step_type)
    self.assertEqual(0.0, timestep.reward)
    self.assertEqual(0.0, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_illegal_state_transition(self):
    self._connection.send = mock.MagicMock(
        return_value=_TERMINATED_STEP_RESPONSE)
    with self.assertRaisesRegex(RuntimeError, 'Environment transitioned'):
      self._env.step({})

  def test_reset(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        side_effect=[_SAMPLE_RESET_RESPONSE, _SAMPLE_STEP_RESPONSE])
    timestep = self._env.reset()

    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)
    self.assertIsNone(timestep.reward)
    self.assertIsNone(timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_spec_generate_value_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    action_spec = self._env.action_spec()
    actions = {
        name: spec.generate_value() for name, spec in action_spec.items()
    }
    self._env.step(actions)
    self._connection.send.assert_called_once_with(
        dm_env_rpc_pb2.StepRequest(
            requested_observations=[1, 2],
            actions={
                1: tensor_utils.pack_tensor(actions['foo']),
                2: tensor_utils.pack_tensor(actions['bar'], dtype=np.str_)
            }))

  def test_reset_changes_spec_raises_error(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        side_effect=[_RESET_CHANGES_SPEC_RESPONSE, _SAMPLE_STEP_RESPONSE])
    with self.assertRaisesRegex(RuntimeError, 'changed spec'):
      self._env.reset()

  def test_observation_spec(self):
    expected_spec = {
        'foo': specs.Array(shape=(), dtype=np.uint8, name='foo'),
        'bar': specs.StringArray(shape=(), name='bar')
    }
    self.assertEqual(expected_spec, self._env.observation_spec())

  def test_action_spec(self):
    expected_spec = {
        'foo': specs.Array(shape=(), dtype=np.uint8, name='foo'),
        'bar': specs.StringArray(shape=(), name='bar')
    }
    self.assertEqual(expected_spec, self._env.action_spec())

  def test_cant_step_after_close(self):
    self._connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.LeaveWorldResponse())
    self._env.close()
    with self.assertRaisesRegex(AttributeError, 'send'):
      self._env.step({})

  def test_reward_spec_default(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.float64), self._env.reward_spec())

  def test_discount_spec_default(self):
    self.assertEqual(
        specs.BoundedArray(
            shape=(), dtype=np.float64, minimum=0.0, maximum=1.0),
        self._env.discount_spec())

  def test_close_leaves_world(self):
    self._connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.LeaveWorldResponse())
    self._env.close()
    self._connection.send.assert_called_once_with(
        dm_env_rpc_pb2.LeaveWorldRequest())

  def test_close_errors_when_cannot_leave_world(self):
    self._connection.send = mock.MagicMock(side_effect=ValueError('foo'))
    with self.assertRaisesRegex(ValueError, 'foo'):
      self._env.close()


class OverrideRewardDiscount(dm_env_adaptor.DmEnvAdaptor):

  def __init__(self):
    self.connection = mock.MagicMock()
    self.reward = mock.MagicMock()
    self.discount = mock.MagicMock()
    super(OverrideRewardDiscount, self).__init__(self.connection, _SAMPLE_SPEC)


class RewardDiscountOverrideTests(absltest.TestCase):

  def test_override_reward(self):
    env = OverrideRewardDiscount()
    env.reward.return_value = 0.5
    env.connection.send.return_value = _SAMPLE_STEP_RESPONSE
    timestep = env.step({})
    self.assertEqual(0.5, timestep.reward)
    env.reward.assert_called()
    self.assertEqual(dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
                     env.reward.call_args[1]['state'])
    self.assertEqual(dm_env.StepType.FIRST,
                     env.reward.call_args[1]['step_type'])
    self.assertDictEqual({
        'foo': 5,
        'bar': 'goodbye'
    }, env.reward.call_args[1]['observations'])

  def test_override_discount(self):
    env = OverrideRewardDiscount()
    env.discount.return_value = 0.5
    env.connection.send.return_value = _SAMPLE_STEP_RESPONSE
    timestep = env.step({})
    self.assertEqual(0.5, timestep.discount)
    env.discount.assert_called()
    self.assertEqual(dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
                     env.discount.call_args[1]['state'])
    self.assertEqual(dm_env.StepType.FIRST,
                     env.discount.call_args[1]['step_type'])
    self.assertDictEqual({
        'foo': 5,
        'bar': 'goodbye'
    }, env.discount.call_args[1]['observations'])


class ReservedKeywordTests(absltest.TestCase):

  def setUp(self):
    super(ReservedKeywordTests, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(self._connection, _RESERVED_SPEC)

  def test_reward_spec(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.uint8), self._env.reward_spec())

  def test_discount_spec(self):
    self.assertEqual(specs.StringArray(shape=()), self._env.discount_spec())

  def test_reward_from_reserved_keyword(self):
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)
    self._env.step({})  # Reward is None for first step.
    timestep = self._env.step({})

    self.assertEqual(5, timestep.reward)

  def test_discount(self):
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)
    timestep = self._env.step({})

    self.assertEqual('goodbye', timestep.discount)

  def test_observations_empty(self):
    self.assertEmpty(self._env.observation_spec())

  def test_explicitly_requesting_reward_and_discount(self):
    env = dm_env_adaptor.DmEnvAdaptor(
        self._connection,
        _RESERVED_SPEC,
        requested_observations=[
            dm_env_adaptor.DEFAULT_REWARD_KEY,
            dm_env_adaptor.DEFAULT_DISCOUNT_KEY
        ])
    expected_observation_spec = {
        dm_env_adaptor.DEFAULT_REWARD_KEY: env.reward_spec(),
        dm_env_adaptor.DEFAULT_DISCOUNT_KEY: env.discount_spec(),
    }
    self.assertEqual(env.observation_spec(), expected_observation_spec)


class EnvironmentAutomaticallyRequestsReservedKeywords(absltest.TestCase):

  def setUp(self):
    super(EnvironmentAutomaticallyRequestsReservedKeywords, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(
        self._connection, _RESERVED_SPEC, requested_observations=[])
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)

  def test_reward_spec_unrequested(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.uint8), self._env.reward_spec())

  def test_discount_spec_unrequested(self):
    self.assertEqual(specs.StringArray(shape=()), self._env.discount_spec())

  def test_does_not_give_back_unrequested_observations(self):
    timestep = self._env.step({})
    self.assertEqual({}, timestep.observation)

  def test_first_reward_none(self):
    timestep = self._env.step({})
    self.assertIsNone(timestep.reward)

  def test_reward_piped_correctly(self):
    self._env.step({})  # Reward is None for first step.
    timestep = self._env.step({})
    self.assertEqual(5, timestep.reward)

  def test_discount_piped_correctly(self):
    timestep = self._env.step({})
    self.assertEqual('goodbye', timestep.discount)


class EnvironmentNestedActionsObservations(absltest.TestCase):

  def test_nested_specs(self):
    env = dm_env_adaptor.DmEnvAdaptor(
        connection=mock.MagicMock(), specs=_SAMPLE_NESTED_SPECS)
    expected_actions = {
        'foo': {
            'bar': specs.Array(shape=(), dtype=np.int32, name='foo.bar'),
        },
        'baz': specs.Array(shape=(), dtype=np.str_, name='baz'),
    }
    expected_observations = {
        'foo': {
            'bar': specs.Array(shape=(), dtype=np.int32, name='foo.bar'),
        },
        'baz': specs.Array(shape=(), dtype=np.str_, name='baz'),
    }

    self.assertSameElements(expected_actions, env.action_spec())
    self.assertSameElements(expected_observations, env.observation_spec())

  def test_no_nested_specs(self):
    env = dm_env_adaptor.DmEnvAdaptor(
        connection=mock.MagicMock(),
        specs=_SAMPLE_NESTED_SPECS,
        nested_tensors=False)
    expected_actions = {
        'foo.bar': specs.Array(shape=(), dtype=np.int32, name='foo.bar'),
        'baz': specs.Array(shape=(), dtype=np.str_, name='baz'),
    }
    expected_observations = {
        'foo.bar': specs.Array(shape=(), dtype=np.int32, name='foo.bar'),
        'baz': specs.Array(shape=(), dtype=np.str_, name='baz'),
    }

    self.assertSameElements(expected_actions, env.action_spec())
    self.assertSameElements(expected_observations, env.observation_spec())

  def test_nested_actions_step(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        return_value=text_format.Parse("""state: RUNNING""",
                                       dm_env_rpc_pb2.StepResponse()))
    env = dm_env_adaptor.DmEnvAdaptor(
        connection, specs=_SAMPLE_NESTED_SPECS, requested_observations=[])

    timestep = env.step({'foo': {'bar': 123}})

    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)

    connection.send.assert_called_once_with(
        text_format.Parse(
            """actions: { key: 1, value: { int32s: { array: 123 } } }""",
            dm_env_rpc_pb2.StepRequest()))

  def test_no_nested_actions_step(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        return_value=text_format.Parse("""state: RUNNING""",
                                       dm_env_rpc_pb2.StepResponse()))
    env = dm_env_adaptor.DmEnvAdaptor(
        connection,
        specs=_SAMPLE_NESTED_SPECS,
        requested_observations=[],
        nested_tensors=False)
    timestep = env.step({'foo.bar': 123})

    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)

    connection.send.assert_called_once_with(
        text_format.Parse(
            """actions: { key: 1, value: { int32s: { array: 123 } } }""",
            dm_env_rpc_pb2.StepRequest()))

  def test_nested_observations_step(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        return_value=text_format.Parse(
            """state: RUNNING
        observations: { key: 1, value: { int32s: { array: 42 } } }""",
            dm_env_rpc_pb2.StepResponse()))

    expected = {'foo': {'bar': 42}}

    env = dm_env_adaptor.DmEnvAdaptor(
        connection,
        specs=_SAMPLE_NESTED_SPECS,
        requested_observations=['foo.bar'])
    timestep = env.step({})
    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)
    self.assertSameElements(expected, timestep.observation)

    connection.send.assert_called_once_with(
        dm_env_rpc_pb2.StepRequest(requested_observations=[1]))

  def test_extensions(self):
    class _ExampleExtension:

      def foo(self):
        return 'bar'

    env = dm_env_adaptor.DmEnvAdaptor(
        connection=mock.MagicMock(),
        specs=_SAMPLE_SPEC,
        extensions={'extension': _ExampleExtension()})

    self.assertEqual('bar', env.extension.foo())

  def test_invalid_extension_attr(self):
    with self.assertRaisesRegex(ValueError,
                                'DmEnvAdaptor already has attribute'):
      dm_env_adaptor.DmEnvAdaptor(
          connection=mock.MagicMock(),
          specs=_SAMPLE_SPEC,
          extensions={'_connection': object()})


class CreateJoinHelpers(absltest.TestCase):

  def test_create_world(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.CreateWorldResponse(
            world_name='Damogran_01'))

    world_name = dm_env_adaptor.create_world(connection, {'planet': 'Damogran'})
    self.assertEqual('Damogran_01', world_name)

    connection.send.assert_called_once_with(
        text_format.Parse(
            """settings: {
                key: 'planet', value: { strings: { array: 'Damogran' } }
            }""", dm_env_rpc_pb2.CreateWorldRequest()))

  def test_join_world(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC))

    env = dm_env_adaptor.join_world(connection, 'Damogran_01',
                                    {'player': 'zaphod'})
    self.assertIsNotNone(env)

    connection.send.assert_called_once_with(
        text_format.Parse(
            """world_name: 'Damogran_01'
                settings: {
                    key: 'player', value: { strings: { array: 'zaphod' } }
                }""", dm_env_rpc_pb2.JoinWorldRequest()))

  def test_create_join_world(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(side_effect=[
        dm_env_rpc_pb2.CreateWorldResponse(world_name='Damogran_01'),
        dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC)
    ])
    env, world_name = dm_env_adaptor.create_and_join_world(
        connection,
        create_world_settings={'planet': 'Damogran'},
        join_world_settings={
            'ship_type': 1,
            'player': 'zaphod',
        })
    self.assertIsNotNone(env)
    self.assertEqual('Damogran_01', world_name)

    connection.send.assert_has_calls([
        mock.call(
            text_format.Parse(
                """settings: {
                key: 'planet', value: { strings: { array: 'Damogran' } }
            }""", dm_env_rpc_pb2.CreateWorldRequest())),
        mock.call(
            text_format.Parse(
                """world_name: 'Damogran_01'
                settings: { key: 'ship_type', value: { int64s: { array: 1 } } }
                settings: {
                    key: 'player', value: { strings: { array: 'zaphod' } }
                }""", dm_env_rpc_pb2.JoinWorldRequest())),
    ])

  def test_create_join_world_with_packed_settings(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(side_effect=[
        dm_env_rpc_pb2.CreateWorldResponse(world_name='Magrathea_02'),
        dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC)
    ])
    env_and_world_name = dm_env_adaptor.create_and_join_world(
        connection,
        create_world_settings={'planet': tensor_utils.pack_tensor('Magrathea')},
        join_world_settings={
            'ship_type': tensor_utils.pack_tensor(2),
            'player': tensor_utils.pack_tensor('arthur'),
            'unpacked_setting': [1, 2, 3],
        })
    self.assertIsNotNone(env_and_world_name.env)
    self.assertEqual('Magrathea_02', env_and_world_name.world_name)

    connection.send.assert_has_calls([
        mock.call(
            text_format.Parse(
                """settings: {
                key: 'planet', value: { strings: { array: 'Magrathea' } }
            }""", dm_env_rpc_pb2.CreateWorldRequest())),
        mock.call(
            text_format.Parse(
                """world_name: 'Magrathea_02'
                settings: { key: 'ship_type', value: { int64s: { array: 2 } } }
                settings: {
                    key: 'player', value: { strings: { array: 'arthur' } }
                }
                settings: {
                    key: 'unpacked_setting', value: {
                      int64s: { array: 1 array: 2 array: 3 }
                      shape: 3
                    }
                }""", dm_env_rpc_pb2.JoinWorldRequest())),
    ])

  def test_create_join_world_with_extension(self):

    class _ExampleExtension:

      def foo(self):
        return 'bar'

    connection = mock.MagicMock()
    connection.send = mock.MagicMock(side_effect=[
        dm_env_rpc_pb2.CreateWorldResponse(world_name='foo'),
        dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC)
    ])

    env, _ = dm_env_adaptor.create_and_join_world(
        connection,
        create_world_settings={},
        join_world_settings={},
        extensions={'extension': _ExampleExtension()})
    self.assertEqual('bar', env.extension.foo())

  def test_create_join_world_with_invalid_extension(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(side_effect=[
        dm_env_rpc_pb2.CreateWorldResponse(world_name='foo'),
        dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC),
        dm_env_rpc_pb2.LeaveWorldResponse(),
        dm_env_rpc_pb2.DestroyWorldRequest()
    ])

    with self.assertRaisesRegex(ValueError,
                                'DmEnvAdaptor already has attribute'):
      _ = dm_env_adaptor.create_and_join_world(
          connection,
          create_world_settings={},
          join_world_settings={},
          extensions={'step': object()})

    connection.send.assert_has_calls([
        mock.call(dm_env_rpc_pb2.CreateWorldRequest()),
        mock.call(dm_env_rpc_pb2.JoinWorldRequest(world_name='foo')),
        mock.call(dm_env_rpc_pb2.LeaveWorldRequest()),
        mock.call(dm_env_rpc_pb2.DestroyWorldRequest(world_name='foo'))
    ])

  def test_created_but_failed_to_join_world(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        side_effect=(
            dm_env_rpc_pb2.CreateWorldResponse(world_name='Damogran_01'),
            error.DmEnvRpcError(status_pb2.Status(message='Failed to Join.')),
            dm_env_rpc_pb2.DestroyWorldResponse()))

    with self.assertRaisesRegex(error.DmEnvRpcError, 'Failed to Join'):
      _ = dm_env_adaptor.create_and_join_world(
          connection, create_world_settings={}, join_world_settings={})

    connection.send.assert_has_calls([
        mock.call(dm_env_rpc_pb2.CreateWorldRequest()),
        mock.call(dm_env_rpc_pb2.JoinWorldRequest(world_name='Damogran_01')),
        mock.call(dm_env_rpc_pb2.DestroyWorldRequest(world_name='Damogran_01'))
    ])

  def test_created_and_joined_but_adaptor_failed(self):
    connection = mock.MagicMock()
    connection.send = mock.MagicMock(
        side_effect=(
            dm_env_rpc_pb2.CreateWorldResponse(world_name='Damogran_01'),
            dm_env_rpc_pb2.JoinWorldResponse(specs=_SAMPLE_SPEC),
            dm_env_rpc_pb2.LeaveWorldResponse(),
            dm_env_rpc_pb2.DestroyWorldResponse()))

    with self.assertRaisesRegex(ValueError, 'Unsupported observations'):
      _ = dm_env_adaptor.create_and_join_world(
          connection,
          create_world_settings={},
          join_world_settings={},
          requested_observations=['invalid_observation'])

    connection.send.assert_has_calls([
        mock.call(dm_env_rpc_pb2.CreateWorldRequest()),
        mock.call(dm_env_rpc_pb2.JoinWorldRequest(world_name='Damogran_01')),
        mock.call(dm_env_rpc_pb2.LeaveWorldRequest()),
        mock.call(dm_env_rpc_pb2.DestroyWorldRequest(world_name='Damogran_01'))
    ])


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Python utilities for flattening and unflattening key-value mappings."""

import collections
from typing import Any, Dict, Mapping


def flatten_dict(input_dict: Mapping[str, Any],
                 separator: str) -> Dict[str, Any]:
  """Flattens mappings by joining sub-keys using the provided separator.

  Only non-empty, mapping types will be flattened. All other types are deemed
  leaf values.

  Args:
    input_dict: Mapping of key-value pairs to flatten.
    separator: Delimiter used to concatenate keys.

  Returns:
    Flattened dictionary of key-value pairs.

  Raises:
    ValueError: If the `input_dict` has a key that contains the separator
      string.
  """
  result = collections.OrderedDict()
  for key, value in input_dict.items():
    if separator in key:
      raise ValueError(f"Key '{key}' already contains separator!")
    if isinstance(value, Mapping) and len(value):
      result.update({
          f'{key}{separator}{sub_key}': sub_value
          for sub_key, sub_value in flatten_dict(value, separator).items()
      })
    else:
      result[key] = value
  return result


def unflatten_dict(input_dict: Mapping[str, Any],
                   separator: str) -> Dict[str, Any]:
  """Unflatten dictionary using split keys to determine the structure.

  For each key, split based on the provided separator and create nested
  dictionary entry for each sub-key.

  Args:
    input_dict: Mapping of key-value pairs to un-flatten.
    separator: Delimiter used to split keys.

  Returns:
    Unflattened dictionary.

  Raises:
    ValueError: If a key, or it's constituent sub-keys already has a value. For
      instance, unflattening `{"foo": True, "foo.bar": "baz"}` will result in
      "foo" being set to both a dict and a Bool.
  """
  result = collections.OrderedDict()
  for key, value in input_dict.items():
    sub_keys = key.split(separator)
    sub_tree = result
    for sub_key in sub_keys[:-1]:
      sub_tree = sub_tree.setdefault(sub_key, {})
      if not isinstance(sub_tree, Mapping):
        raise ValueError(f"Sub-tree '{sub_key}' has already been assigned a "
                         "leaf value {sub_tree}")

    if sub_keys[-1] in sub_tree:
      raise ValueError(f'Duplicate key {key}')
    sub_tree[sub_keys[-1]] = value
  return result
